/*************************************************************
 *                 bbomberman                                *
 *  (c) 2004 by Bernhard Eder (bb-generation)                *
 *  http://bbgen.net/bomberman for updates                   *
 *  Webmaster@bbgen.net for bug reports and other stuff      *
 * --------------------------------------------------------- *
 * This program is free software; you can modify it under    *
 * the terms of the GNU General Public License as published  *
 * by the Free Software Foundation, located in LICENSE       *
 *                                                           *
 *************************************************************/
 
#include "cl_server.h"
/* FIXME: debug */
#include <iostream>


/*
 * cl_bomb implementation
 */

//ok ramblers, let's get rambling
cl_bomb::cl_bomb(cl_bomberman& bomb_owner) :
  owner(bomb_owner), x(bomb_owner.x), y(bomb_owner.y)
{
  _M_creation_tick = bomb_owner.server.tick_count();
}

/* protected constructor for inheritance by cl_bomb_remote */
cl_bomb::cl_bomb(cl_bomberman& bomb_owner, const int x, const int y, const long long creation_tick) :
  owner(bomb_owner), x(x), y(y), _M_creation_tick(creation_tick)
{ }

/* access creation tick (read only) */
long long cl_bomb::creation_tick() const
{
  return _M_creation_tick;
}

/* set new creation tick */
void cl_bomb::creation_tick(const long long new_creation_tick)
{
  /* FIXME: net; somehow notify client..?
   *   (strictly speaking, this isn't necessary:
   *    bombs are only exploded on the server, the appropriate dynamic fields
   *    will be sent to the client no matter what his local cl_bomb copy's
   *    _M_creation_tick is. Furthermore, cl_game draws dynamic_fields after
   *    movable objects, so the only difference will be an unexploded bomb
   *    hidden below an explosion dynamic field. ;) )
   */
  _M_creation_tick = new_creation_tick;
}

/* explode the bomb (after it has just been ticking) */
void cl_bomb::explode()
{
  /* explosion starts at the bomb itself... */
  //std::cerr << "* cl_bomb::explode() exploding bomb field itself" << std::endl;
  explode_field(x, y, BB_DIRECTION_NONE);

  /* ...and continues into all 4 directions */
  //std::cerr << "* cl_bomb::explode() spreading explosion" << std::endl;
  for (int dir = 0; dir < 4; dir++)
    explode_to(bb_direction(dir));
}

/* explode a single field */
bool cl_bomb::explode_field(const int x, const int y, const bb_direction& direction)
{
  //std::cerr << "* cl_bomb::explode_field(" << x << ", " << y << ", " << direction << ")" << std::endl;
  bool        was_box = false;
  cl_server&  server = owner.server;

  /* stop if target field is out of map */
  if ((x < 0) || (y < 0) ||
      (x >= server.map_fields_x()) || (y >= server.map_fields_y()))
    return false;

  const bb_map_field  map_field = server.map(x, y);

  /* stop at fixed box */
  if (map_field == BB_MAP_FIELD_BOX_FIXED)
    return false;

  /* is target field a normal box? */
  if (map_field == BB_MAP_FIELD_BOX)
    was_box = true;

  server.explode_field(x, y, &owner);

  /* check every bomberman's position */
  for (cl_server::bombermans_type::iterator bman = server.bombermans.begin();
       bman < server.bombermans.end(); bman++)
  {
    /* check for bomb at this field */
    for (cl_bomberman::bombs_type::iterator bomb = (*bman)->bombs.begin();
         bomb < (*bman)->bombs.end(); bomb++)
    {
      if (((*bomb)->x == x) && ((*bomb)->y == y) && (&(**bomb) != this))
        /* explode it now, too */
        /* FIXME: this won't work if this bomb is past the bomb currently
         *   examined in server->bombs... it'll be examined by server->tick()
         *   at next tick again, but then it'll already have
         *   status() == BB_BOMB_STAT_VANISHED... :(
         */
        (*bomb)->creation_tick(_M_creation_tick); //chain reaction
    }
  }

  if (was_box)
  {
    /* FIXME: the following code will prevent any box from being destroyed
     *   if the bomb wasn't set directly beneath it O_o
     *   Is this effect desired? bb..?
     *
     *   It indead is, since the bomb is exploded three times...
     *     (zzz 2004-10-09)
     */
    //std::cerr << "* cl_bomb::explode_field() direction examinations..." << std::endl;
    if ((direction == BB_DIRECTION_LEFT) &&
        (x - 1 >= 0) &&
        (server.map(x - 1, y) != BB_MAP_FIELD_EMPTY))
      return false;
    else if ((direction == BB_DIRECTION_RIGHT) &&
             (x + 1 < server.map_fields_x()) &&
             (server.map(x + 1, y) != BB_MAP_FIELD_EMPTY))
      return false;
    else if ((direction == BB_DIRECTION_UP) &&
             (y - 1 >= 0) &&
             (server.map(x, y - 1) != BB_MAP_FIELD_EMPTY))
      return false;
    else if ((direction == BB_DIRECTION_DOWN) &&
             (y + 1 < server.map_fields_y()) &&
             (server.map(x, y + 1) != BB_MAP_FIELD_EMPTY))
      return false;
  }

  //std::cerr << "* cl_bomb::explode_field() returning true" << std::endl;
  return true;
}

/* continue explosion into one of the 4 directions */
bool cl_bomb::explode_to(const bb_direction& direction)
{
  /* continue explosion for expforce fields */
  for (int i = 1; i <= owner.expforce; i++)
  {
    /* continue in which direction? */
    switch (direction)
    {
      case BB_DIRECTION_LEFT:
        if (!explode_field(x - i, y, direction))
          return false;
        break;
      case BB_DIRECTION_RIGHT:
        if (!explode_field(x + i, y, direction))
          return false;
        break;
      case BB_DIRECTION_UP:
        if (!explode_field(x, y - i,direction))
          return false;
        break;
      case BB_DIRECTION_DOWN:
        if (!explode_field(x, y + i, direction))
          return false;
        break;
      default:
        throw std::runtime_error("cl_bomb::explode_to(): invalid direction");
    }
  }

  /* indicate explosion was only stopped by expforce limit */
  return true;
}

/* determine explosion status of bomb */
bb_bomb_status cl_bomb::status() const
{
  const long long tick_count = owner.server.tick_count();
  
  if (tick_count < _M_creation_tick + 20)
    return BB_BOMB_STAT_TICKING1;
  else if (tick_count < _M_creation_tick + 30)
    return BB_BOMB_STAT_TICKING2;
  /* FIXME: whatsthis; so every bomb should explode three times..? O_o
   * 
   *   It indead should... so that the user sees smooth explosion continuation
   *   (just guessing, though...)  (zzz 2004-10-09)
   */
  else if (tick_count < _M_creation_tick + 33)
    return BB_BOMB_STAT_EXPLODING;  // BBOOOOMM, BOOOMMM, every time you drop the bomb..
  else
    return BB_BOMB_STAT_VANISHED;
}


/*
 * cl_bomberman implementation
 */

/* constructor */
cl_bomberman::cl_bomberman(cl_server& server, const int id, const std::string& name, cl_client* client) :
  server(server), id(id), name(name), client(client)
{
  /* initialize member vars that have constant start value */
  alive = true;
  bombs_max = 1;
  direction = BB_DIRECTION_DOWN;
  expforce = 1;

  /* initialize coordinates from id */
  if (id % 2 == 0)
    y = 0;
  else
    y = server.map_fields_y() - 1;

  if (id == 0 || id == 3)
    x = 0;
  else
    x = server.map_fields_x() - 1;
}

/* protected constructor for inheritance by cl_bomberman_remote */
cl_bomberman::cl_bomberman(cl_server& server, const int id, const std::string& name, const bool alive, const int x, const int y, const bb_direction direction, const int bombs_max, const int expforce, cl_client* client) :
  server(server), id(id), name(name), alive(alive), x(x), y(y), direction(direction), bombs_max(bombs_max), expforce(expforce), client(client)
{ }

/* kill the bomberman */
void cl_bomberman::do_kill(const cl_bomberman* killer)
{
  /* FIXME: invalidate client object reference to forcefully prevent user
   *   from further interactions with the server..? 
   */
  killed_by = killer;
  alive = false;
}

/* really move bomberman */
bool cl_bomberman::do_move(const bb_direction& to)
{
  int to_x = x, to_y = y;

  /* calculate destination coordinates */
  switch (to)
  {
    case BB_DIRECTION_LEFT:
      to_x--;
      break;
    case BB_DIRECTION_RIGHT:
      to_x++;
      break;
    case BB_DIRECTION_UP:
      to_y--;
      break;
    case BB_DIRECTION_DOWN:
      to_y++;
      break;
    default:
      throw std::runtime_error("cl_game::do_move_bomberman(): invalid direction");
  }

  /* turn the bomberman to specified direction */
  direction = to;

  /* can the bomberman be moved to destination? */
  if (!server.is_field_passable(to_x, to_y))
    return false;

  /* move to destination */
  x = to_x; y = to_y;

  /* indicate successful move */
  return true;
}

/* place a bomb */
bool cl_bomberman::do_place_bomb()
{
  /* maximum simultaneous bombs placed? */
  if ((bombs_max < 0) || (bombs.size() >= (unsigned int) bombs_max))
    return false;

  /* place bomb */
  bombs.push_back(bomb_ptr(new cl_bomb(*this)));

  /* indicate successful bomb placement */
  return true;
}

/* process a powerup (from dynamic field in server) */
void cl_bomberman::do_powerup(const bb_dynamic_field powerup)
{
  switch (powerup)
  {
    case BB_DYN_FIELD_PUP_BOMBSMAX:
      /* concurrent bomb count powerup */
      bombs_max++;
      break;
    case BB_DYN_FIELD_PUP_EXPFORCE:
      /* expforce powerup */
      expforce++;
      break;
    default:
      throw std::runtime_error("cl_bomberman::do_powerup(): invalid powerup");
  }
}

/* move bomberman (only queues message for associated client) */
void cl_bomberman::move(const bb_direction& to) const
{
  if (!client)
    throw std::runtime_error("cl_bomberman::move(): no client attached");

  cl_proto_msg msg(BB_PROTO_MSG_DO_BBMOVE);

  msg.push_arg(id);
  msg.push_arg(to);
  client->to_server.queue_msg(msg);
}

/* place bomb (only queues message for associated client) */
void cl_bomberman::place_bomb() const
{
  if (!client)
    throw std::runtime_error("cl_bomberman::move(): no client attached");

  cl_proto_msg msg(BB_PROTO_MSG_DO_BBBOMB);

  msg.push_arg(id);
  client->to_server.queue_msg(msg);
}


/*
 * cl_client implementation
 */

/* constructor */
cl_client::cl_client(cl_server& server) :
  kind(BB_CLIENT_LOCAL), server(server)
{ }

cl_client::cl_client(bb_client_kind kind, cl_server& server) :
  kind(kind), server(server)
{ }

/* FIXME: automagically call from destructor? maybe only unless called before? */
/* FIXME: queue connection-aborted message for client? */
/* FIXME: where detach() is used, usually clients.erase() follows;
 *   maybe don't be so strict, but let the client get a chance to process a
 *   connection-aborted / error message..?
 */
/* detach client from server */
void cl_client::detach()
{
  std::cerr << "* cl_client::detach()" << std::endl;

  /* iterate through server's bombermans */
  for (cl_server::bombermans_type::iterator  bman = server.bombermans.begin();
       bman < server.bombermans.end(); bman++)
  {
    /* is bomberman attached to this client? */
    if ((*bman)->client == this)
    {
      /* detach from client */
      (*bman)->client = NULL;
    }
  }
}


/*
 * cl_server virtual class implementation
 *   code common to every server (local, local with net, remote)
 */

/* access bomberman */
cl_bomberman& cl_server::bomberman(const int id) const
{
  /* is requested id valid? */
  if ((id < 0) || (id >= (const int) bombermans.size()))
    throw std::runtime_error("cl_server::bomberman(): invalid id");

  return *bombermans[id];
}

/* access dynamic fields (read only) */
cl_server::dynamic_fields_type cl_server::dynamic_fields() const
{
  return dynamic_fields_type(_M_dynamic_fields);
}

/* access dynamic fields at specific coordinates (read only) */
cl_server::dynamic_fields_type cl_server::dynamic_fields(const int x, const int y) const
{
  bb_map_field_xy  xy(x, y);

  /* retrieve dynamic fields for bomberman's position */
  std::pair<_dyn_fs_const_iterator, _dyn_fs_const_iterator>
    iterators = _M_dynamic_fields.equal_range(xy);

  /* no such fields && neither any "greater" ones? */
  if (iterators.first == _M_dynamic_fields.end())
    return dynamic_fields_type();

  /* did multimap return "greater" fields? */
  if (iterators.first->first != xy)
    return dynamic_fields_type();

  /* really fields found; return multimap containing copies of them */
  return dynamic_fields_type(iterators.first, iterators.second);
}

/* determine field passability (for bombermans) */
bool cl_server::is_field_passable(const int x, const int y) const
{
  /* is field outside of map? */
  if ((x < 0) || (y < 0) ||
      (x >= _M_map_fields_x) || (y >= _M_map_fields_y))
    return false;

  /* is a box at this field? */
  if (_M_map[x][y] == BB_MAP_FIELD_BOX ||
      _M_map[x][y] == BB_MAP_FIELD_BOX_FIXED)
    return false;
  
  /* is a player or an unexploded bomb at this field? */
  for (bombermans_type::const_iterator bman = bombermans.begin();
       bman < bombermans.end(); bman++)
  {
    if (((*bman)->x == x) && ((*bman)->y == y))
      return false;

    for (cl_bomberman::bombs_type::const_iterator bomb = (*bman)->bombs.begin();
         bomb < (*bman)->bombs.end(); bomb++)
    {
      if (((*bomb)->x == x) && ((*bomb)->y == y))
        return false;
    }
  }
  
  /* field is passable */
  return true;
}

bb_map_field cl_server::map(const int x, const int y) const
{
  return _M_map[x][y];
}

int cl_server::map_fields_x() const
{
  return _M_map_fields_x;
}

int cl_server::map_fields_y() const
{
  return _M_map_fields_y;
}

long long cl_server::tick_count() const
{
  return _M_ticks;
}

/* create a dynamic field */
void cl_server::_M_create_dynamic_field(const int x, const int y, const bb_dynamic_field field_kind)
{
  std::cerr << "* cl_server::_M_create_dynamic_field(" << x << "; " << y << "; " << field_kind << ")" << std::endl;

  _M_dynamic_fields.insert(std::make_pair(bb_map_field_xy(x, y), field_kind));
}

/* destroy a single dynamic field */
void cl_server::_M_destroy_dynamic_field(const int x, const int y, const bb_dynamic_field field_kind)
{
  bb_map_field_xy  xy(x, y);

  /* iterate through all dynamic fields at these coordinates */
  for (dynamic_fields_type::iterator  field = _M_dynamic_fields.find(xy);
       (field != _M_dynamic_fields.end()) && (field->first == xy); field++)
  {
    /* field of specified kind found? */
    if (field->second == field_kind)
    {
      /* delete field & return (only one field was to be deleted) */
      _M_dynamic_fields.erase(field);
      return;
    }
  }

  throw std::runtime_error("cl_server::_M_destroy_dynamic_field(): invalid field");
}

/* encode a map column (all field with same x coordinate) as string
 * (for transport over network)
 */
std::string cl_server::_M_encode_map_column(const int x) const
{
  std::string                       out;
  const std::vector<bb_map_field>&  column = _M_map[x];

  /* iterate through fields in map column */
  for (int y = 0; y < _M_map_fields_y; y++)
  {
    /* examine field, encode as string & append to output */
    switch (column[y])
    {
      case BB_MAP_FIELD_EMPTY:
        out += 'E';
        break;
      case BB_MAP_FIELD_BOX:
        out += 'B';
        break;
      case BB_MAP_FIELD_BOX_FIXED:
        out += 'F';
        break;
      default:
        throw std::runtime_error("cl_server::_M_encode_map_column(): invalid map field");
    }
  }

  return out;
}

/* decode map column from string (received from network) */
void cl_server::_M_decode_map_column(const std::string& in)
{
  /* map creation already begun? */
  if (_M_map.empty())
  {
    /* begin of map creation; initialize map members */
    _M_map_fields_x = 0;
    _M_map_fields_y = in.size();
  }
  else
  {
    /* ensure each column is of same size */
    if ((int) in.size() != _M_map_fields_y)
      throw std::runtime_error("cl_server::_M_decode_map_column(): input column doesn't match map size");
  }

  /* decode map column */
  std::vector<bb_map_field>  column(_M_map_fields_y);
  for (int y = 0; y < _M_map_fields_y; y++)
  {
    switch (in[y])
    {
      case 'E':
        column[y] = BB_MAP_FIELD_EMPTY;
        break;
      case 'B':
        column[y] = BB_MAP_FIELD_BOX;
        break;
      case 'F':
        column[y] = BB_MAP_FIELD_BOX_FIXED;
        break;
      default:
        throw std::runtime_error("cl_server::_M_decode_map_column(): invalid map field");
    }
  }

  /* FIXME: this is _really_ bad memory fragmentation... std::vector will
   *   realloc new memory for the pseudo array each time a map column is
   *   added... But at the moment, having only a single map-block-transfer
   *   message is just too tempting... ;)
   *
   *   (Additionally, the whole column will be copied when added to the
   *    vector.. but that's just how std::vectors work. (They take const
   *    references as arguments to push_back() etc.))
   *
   *   (zzz 2004-10-11)
   */
  /* add column to map */
  _M_map.push_back(column);
  _M_map_fields_x++;
}


/*
 * cl_server_local implementation
 *   local server specific code
 */

/* constructor */
cl_server_local::cl_server_local(const int map_fields_x, const int map_fields_y, const bomberman_names_type& bomberman_names)
{
  std::cerr << "+ cl_server_local::cl_server_local(...)" << std::endl;
  std::cerr << "  map_fields_x: " << map_fields_x << "; map_fields_y: " << map_fields_y << std::endl;

  /* FIXME: language */
#if 0
  /* initialize member vars */
  this->map_fields_x = map_fields_x;
  this->map_fields_y = map_fields_y;
#endif
  _M_map_fields_x = map_fields_x;
  _M_map_fields_y = map_fields_y;

  /* initialize tick counter */
  _M_ticks = -1;
  
  /* create a local client */
  std::cerr << "* cl_server_local::cl_server_local(...) creating local client" << std::endl;
  client_ptr  client(new cl_client(*this));
  clients.push_back(client);
  
  /* creat one bomberman for each string in bomberman_names */
  std::cerr << "* cl_server_local::cl_server_local(...) creating bombermans" << std::endl;
  int id = 0;
  for (bomberman_names_type::const_iterator bman_name = bomberman_names.begin();
       bman_name < bomberman_names.end(); bman_name++)
  {
    /* create a bomberman attached to local client */
    std::cerr << "* cl_server_local::cl_server_local(...) creating a bomberman" << std::endl;
    bombermans.push_back(bomberman_ptr(
      new cl_bomberman(*this, id, *bman_name, client.get())
    ));

    /* increase id counter */
    id++;
  }

  /* build a map */
  std::cerr << "* cl_server_local::cl_server_local(...) building map" << std::endl;
  _M_build_map();

  std::cerr << "- cl_server_local::cl_server_local(...)" << std::endl;
}

/* destructor */
cl_server_local::~cl_server_local()
{
  bombermans.clear();
  clients.clear();
  /* FIXME: causes compiler crash; CAUTION!! Withouth ulimit on VSZ,
   *   compiling this line will let your machine's virtual memory be exhausted!
   *   The kernel will kill random processes, before it finally kills cc1plus!
   *   (The information above is valid for:
   *    GNU C++ version 3.3.4 (Debian 1:3.3.4-6sarge1) (i486-linux)
   *            compiled by GNU C version 3.3.4 (Debian 1:3.3.4-6sarge1).)
   *   This is probably due to the std::vector<std::vector<int> > definition
   *   of map, chaining two objects with overloaded functions.
   */
  //map.clear;
}

/* explode a map field */
void cl_server_local::explode_field(const int x, const int y, const cl_bomberman* causer)
{
  std::cerr << "* cl_server_local::explode_field(" << x << "; " << y << ")" << std::endl;

  /* throw exception if target field is out of map */
  if ((x < 0) || (y < 0) ||
      (x >= _M_map_fields_x) || (y >= _M_map_fields_y))
    throw std::runtime_error("cl_server_local::explode_field(): coordinates out of map");

  const bb_map_field  map_field = _M_map[x][y];

  /* throw exception if trying to explode a fixed box */
  if (map_field == BB_MAP_FIELD_BOX_FIXED)
    throw std::runtime_error("cl_server_local::explode_field(): attempt to explode fixed box");

  /* FIXME: int rnd = rand() % 6; */
  /* is target field a normal box? */
  if (map_field == BB_MAP_FIELD_BOX)
  {
    /* set map field to empty */
    _M_set_map_field(x, y, BB_MAP_FIELD_EMPTY);

    /* does pseudo-random number generator suppose to place a powerup? */
    if (!(rand() % 8))
    {
      std::cerr << "* cl_server_local::explode_field() creating powerup" << std::endl;
      
      /* which type of powerup? */
      if (rand() % 2)
        _M_create_dynamic_field(x, y, BB_DYN_FIELD_PUP_BOMBSMAX);
      else
        _M_create_dynamic_field(x, y, BB_DYN_FIELD_PUP_EXPFORCE);
    }
  }

  /* check every bomberman's position */
  for (bombermans_type::iterator bman = bombermans.begin();
       bman < bombermans.end(); bman++)
  {
    /* kill bomberman if at target field (and still alive) */
    if (((*bman)->alive) && ((*bman)->x == x) && ((*bman)->y == y))
    {
      std::cerr << "* cl_server_local::explode_field() killing bomberman" << std::endl;
      (*bman)->do_kill(causer);  //..you kill the god your child has born

      /* notify all clients */
      cl_proto_msg  msg(BB_PROTO_MSG_EV_BBKILL);
      msg.push_arg((*bman)->id);
      msg.push_arg(causer->id);
      _M_queue_for_clients(msg);
    }
  }

  /* FIXME: here SDL drawing would take place */
  /* create a dynamic field for explosion */
  _M_create_dynamic_field(x, y, BB_DYN_FIELD_EXPLOSION);
}

void cl_server_local::tick()
{
  /* increase tick counter */
  _M_ticks++;
  cl_proto_msg  msg(BB_PROTO_MSG_EV_TICK);
  msg.push_arg(_M_ticks);
  _M_queue_for_clients(msg);

  /* destroy explosion dynamic fields (from previous tick) */
  _M_destroy_dynamic_fields(BB_DYN_FIELD_EXPLOSION);

  /* process messages in client queues
   * (one message per tick & client)
   */
  _M_process_client_msgs();

  /* process bombs -- cause bomb explosions etc. */
  _M_process_bombs();
}

/* check for & process powerups for bomberman */
void cl_server_local::_M_bomberman_powerup(cl_bomberman& bomberman)
{
  bb_map_field_xy  xy(bomberman.x, bomberman.y);

  /* retrieve dynamic fields for bomberman's position */
  std::pair<_dyn_fs_iterator, _dyn_fs_iterator>
    iterators = _M_dynamic_fields.equal_range(xy);

  /* no such fields && neither any "greater" ones? */
  if (iterators.first == _M_dynamic_fields.end())
    return;

  /* did multimap return "greater" fields? */
  if (iterators.first->first != xy)
    return;

  /* iterate through range */
  while (iterators.first != iterators.second)
  {
    bb_dynamic_field  dynamic_field = iterators.first->second;
    cl_proto_msg      msg(BB_PROTO_MSG_EV_BBPUP);

    /* is dynamic field a powerup? */
    switch (dynamic_field)
    {
      case BB_DYN_FIELD_PUP_BOMBSMAX:
      case BB_DYN_FIELD_PUP_EXPFORCE:
        /* pass powerup on to bomberman */
        bomberman.do_powerup(dynamic_field);

        /* notify all clients */
        msg.push_arg(bomberman.id);
        msg.push_arg(dynamic_field);
        _M_queue_for_clients(msg);

        /* delete dynamic field */
        _M_destroy_dynamic_field(iterators.first);
        break;
      default:
        /* no powerup */
        break;
    }

    iterators.first++;
  }
}

void cl_server_local::_M_build_map()
{
  /* resize map vector */
  /* old form was short-hand, using implicit casts:
   *   map.assign(map_fields_x, map_fields_y);
   * new one shows what really happens
   * FIXME: revert back to easier-to-pseudo-understand implicit form later..?
   *   (the implicit form would _not_ work if it'd be somewhat like a
   *    three dimensional vector (cascade))
   */
  _M_map.assign(_M_map_fields_x, std::vector<bb_map_field>(_M_map_fields_y, BB_MAP_FIELD_EMPTY));

  for (int i = 0; i < _M_map_fields_y; i++)
  {
    if ((i == 0) || i == (_M_map_fields_y - 1))  // first and last row
    {
      /* FIXME */
#if 1
      _M_map[0][i]
        = _M_map[1][i]
        = _M_map[_M_map_fields_x - 1][i]
        = _M_map[_M_map_fields_x - 2][i]
        = BB_MAP_FIELD_EMPTY;
#else
      _M_set_map_field(0, i, BB_MAP_FIELD_EMPTY);
      _M_set_map_field(1, i, BB_MAP_FIELD_EMPTY);
      _M_set_map_field(_M_map_fields_x - 1, i, BB_MAP_FIELD_EMPTY);
      _M_set_map_field(_M_map_fields_x - 2, i, BB_MAP_FIELD_EMPTY);
#endif

      for (int k = 2; k < _M_map_fields_x - 2; k++)
#if 1
        _M_map[k][i] = BB_MAP_FIELD_BOX;
#else
        _M_set_map_field(k, i, BB_MAP_FIELD_BOX);
#endif
    }
    else
    {
      if (i % 2 != 0)
      {
        for (int k = 0; k < _M_map_fields_x; k++)
        {
          if (k % 2 == 0)
          {
            if (std::rand() % 5 == 0)
#if 1
              _M_map[k][i] = BB_MAP_FIELD_EMPTY;
#else
              _M_set_map_field(k, i, BB_MAP_FIELD_EMPTY);
#endif
            else
#if 1
              _M_map[k][i] = BB_MAP_FIELD_BOX;
#else
              _M_set_map_field(k, i, BB_MAP_FIELD_BOX);
#endif
          }
          else
#if 1
            _M_map[k][i] = BB_MAP_FIELD_BOX_FIXED;
#else
            _M_set_map_field(k, i, BB_MAP_FIELD_BOX_FIXED);
#endif
        }
      }
      else
      {
        for (int k = 0; k < _M_map_fields_x; k++)
#if 1
          _M_map[k][i] = BB_MAP_FIELD_BOX;
#else
          _M_set_map_field(k, i, BB_MAP_FIELD_BOX);
#endif
      }
    }

    if ((i == 1) || (i == _M_map_fields_y - 2))  // second and the one before the last row
    {
      /* FIXME */
#if 1
      _M_map[0][i]
        = _M_map[_M_map_fields_x - 1][i]
        = BB_MAP_FIELD_EMPTY;
#else
      _M_set_map_field(0, i, BB_MAP_FIELD_EMPTY);
      _M_set_map_field(_M_map_fields_x - 1, i, BB_MAP_FIELD_EMPTY);
#endif
    }
  }
}

void cl_server_local::_M_create_dynamic_field(const int x, const int y, const bb_dynamic_field field_kind)
{
  /* create dynamic field using virtual function inherited from cl_server */
  cl_server::_M_create_dynamic_field(x, y, field_kind);

  /* notify all clients */
  cl_proto_msg  msg(BB_PROTO_MSG_EV_DFCREATE);
  msg.push_arg(x);
  msg.push_arg(y);
  msg.push_arg(field_kind);
  _M_queue_for_clients(msg);
}

/* destroy a single dynamic field */
void cl_server_local::_M_destroy_dynamic_field(const dynamic_fields_type::iterator& field)
{
  /* notify all clients */
  cl_proto_msg  msg(BB_PROTO_MSG_EV_DFERASE);
  msg.push_arg(field->first.x);
  msg.push_arg(field->first.y);
  msg.push_arg(field->second);
  _M_queue_for_clients(msg);

  /* destroy dynamic field by referencing it with an iterator */
  _M_dynamic_fields.erase(field);
}

/* destroy a single dynamic field */
void cl_server_local::_M_destroy_dynamic_field(const int x, const int y, const bb_dynamic_field field_kind)
{
  /* notify all clients */
  cl_proto_msg  msg(BB_PROTO_MSG_EV_DFERASE);
  msg.push_arg(x);
  msg.push_arg(y);
  msg.push_arg(field_kind);
  _M_queue_for_clients(msg);

  /* destroy dynamic field using virtual function inherited from cl_server */
  cl_server::_M_destroy_dynamic_field(x, y, field_kind);
}

/* destroy selected kind of dynamic fields */
void cl_server_local::_M_destroy_dynamic_fields(const bb_dynamic_field field_kind)
{
  /* iterate through all dynamic fields */
  for (_dyn_fs_iterator dynamic_field = _M_dynamic_fields.begin();
       dynamic_field != _M_dynamic_fields.end(); dynamic_field++)
  {
    /* is the field of selected kind? */
    if (dynamic_field->second == field_kind)
    {
      /* notify all clients */
      cl_proto_msg  msg(BB_PROTO_MSG_EV_DFERASE);
      msg.push_arg(dynamic_field->first.x);
      msg.push_arg(dynamic_field->first.y);
      msg.push_arg(dynamic_field->second);
      _M_queue_for_clients(msg);
      
      /* destroy dynamic field */
      _M_dynamic_fields.erase(dynamic_field);
    }
  }
}

/* process bombs -- cause bomb explosions etc. */
void cl_server_local::_M_process_bombs()
{
  cl_proto_msg  msg(BB_PROTO_MSG_EV_DELBOMB);

  for (bombermans_type::iterator bman = bombermans.begin();
       bman < bombermans.end(); bman++)
  {
    for (cl_bomberman::bombs_type::iterator bomb = (*bman)->bombs.begin();
         bomb < (*bman)->bombs.end(); bomb++)
    {
      switch ((*bomb)->status())
      {
        case BB_BOMB_STAT_TICKING1:
        case BB_BOMB_STAT_TICKING2:
          break;

        case BB_BOMB_STAT_EXPLODING:
          /* explode bomb; spreads to nearby fields */
          std::cerr << "* cl_server_local::tick() exploding bomb" << std::endl;
          (*bomb)->explode();
          break;

        case BB_BOMB_STAT_VANISHED:
          /*
          delete bomb;
          player[i]->bombs.erase(std::remove(player[i]->bombs.begin(),player[i]->bombs.end(),player[i]->bombs[j]),player[i]->bombs.end());
          */
          /* notify clients to delete their copies of this bomb */
          msg.clear_args();
          msg.push_arg((*bman)->id);
          /* FIXME: add id field to cl_bomb */
#if 0
          msg.push_arg((*bomb)->id);
#endif
          _M_queue_for_clients(msg);

          /* delete exploded & vanished bomb */
          std::cerr << "* cl_server_local::tick() erasing bomb" << std::endl;
          (*bman)->bombs.erase(bomb);
          break;

        default:
          throw std::runtime_error("cl_server_local::tick(): cl_bomb::status() returned invalid value");
      }
    }
  }
}

/* process messages in client queues */
void cl_server_local::_M_process_client_msgs()
{
  for (clients_type::iterator client = clients.begin();
       client < clients.end(); client++)
  {
    if ((*client)->to_server.empty())
      continue;

    cl_proto_msg   msg = (*client)->to_server.poll_msg();
    cl_bomberman*  bman;

    try
    {
      /* examine what action to take */
      switch (msg.type())
      {
        case BB_PROTO_MSG_DO_BBMOVE:
          /* move indicated bomberman */
          bman = &bomberman(msg.arg_to_int(0));
          if (bman->do_move((bb_direction) msg.arg_to_int(1)))
          {
            /* successful move; notify all clients */
            _M_queue_for_clients(cl_proto_msg(BB_PROTO_MSG_EV_BBMOVE, msg));

            /* check for powerups */
            _M_bomberman_powerup(*bman);
          }
          else
          {
            /* move denied; notify only sending client */
            (*client)->to_client.queue_msg(cl_proto_msg(BB_PROTO_MSG_DENY_BBMOVE, msg));
          }
          break;

        case BB_PROTO_MSG_DO_BBBOMB:
          bman = &bomberman(msg.arg_to_int(0));
          if (bman->do_place_bomb())
          {
            /* successful bomb placement; notify all clients */
            _M_queue_for_clients(cl_proto_msg(BB_PROTO_MSG_EV_BBBOMB, msg));
          }
          else
          {
            /* bomb placement denied; notify only sending client */
            (*client)->to_client.queue_msg(cl_proto_msg(BB_PROTO_MSG_DENY_BBBOMB, msg));
          }
          break;

        default:
          throw std::runtime_error("cl_server_local::tick(): invalid message type");
      }
    }
    catch (std::runtime_error err)
    {
      /* client message caused an exception -- detach client */
      (*client)->detach();
      clients.erase(client);
    }
  }
}

/* queue a network protocol message for all clients */
void cl_server_local::_M_queue_for_clients(const cl_proto_msg& msg)
{
  for (clients_type::iterator  client = clients.begin();
       client < clients.end(); client++)
  {
    (*client)->to_client.queue_msg(msg);
  }
}

/* set map field & notify all clients */
void cl_server_local::_M_set_map_field(const int x, const int y, const bb_map_field field_kind)
{
  _M_map[x][y] = field_kind;

  cl_proto_msg  msg(BB_PROTO_MSG_EV_MAPFIELD);
  msg.push_arg(x);
  msg.push_arg(y);
  msg.push_arg(field_kind);

  _M_queue_for_clients(msg);
}
